import cntk as C
from cntkx.layers import QRNN
from cntk.layers import Dense, Recurrence, LSTM
import numpy as np
import random
import time


def generate_variable_10(nb_samples=1000, dim=3):
    """
    Generate a dataset of sequences where sequence length is between 5 to 20. Sequence will have values 1 or 0.
    Target will be the sum of the sequence.
    """
    r = [random.randint(5, 20) for __ in range(nb_samples)]
    x = [np.random.randint(2, size=(t, dim)).astype(np.float32) for t in r]
    y = np.array([arr.sum() for arr in x])[..., None]
    return x, y


input_dim = 2
x, y = generate_variable_10(nb_samples=1000, dim=input_dim)

hidden_dim = 100

input_tensor = C.sequence.input_variable(input_dim)
target_tensor = C.input_variable(1)

hidden = QRNN(window=2, hidden_dim=hidden_dim)(input_tensor)
# hidden = Recurrence(LSTM(hidden_dim))(input_tensor)
prediction = Dense(1)(C.sequence.last(hidden))

loss = C.squared_error(prediction, target_tensor)
sgd_m = C.momentum_sgd(prediction.parameters, 0.1, 0.912)
trainer = C.Trainer(prediction, (loss,), [sgd_m])

n_epoch = 50
minibatch_size = 30

start = time.time()
for epoch in range(n_epoch):

    for i in range(0, len(x), minibatch_size):
        lbound, ubound = i, i + minibatch_size
        x_mini = x[lbound:ubound]
        y_mini = y[lbound:ubound]
        trainer.train_minibatch({input_tensor: x_mini,
                                 target_tensor: y_mini})

        print(f"loss: {trainer.previous_minibatch_loss_average}")

print(f"Training completed in {time.time() - start}s")
n = np.random.randint(2, size=(6, input_dim)).astype(np.float32)
print(prediction.eval({input_tensor: [n]}))
print(n.sum())



import cntk as C
import cntkx as Cx
import numpy as np
from cntk.layers import Dropout, MaxPooling, BatchNormalization, AveragePooling
from cntk.layers.blocks import identity
from cntkx.layers import Conv2DMaxPool, Dense, Convolution2D


def VGG16(num_classes: int):
    """ for image classification """

    layer1 = Conv2DMaxPool(2, conv_filter_shape=(3, 3), pool_filter_shape=(2, 2), conv_num_filters=64,
                           activation=C.relu, conv_pad=True, pool_strides=(2, 2), name_prefix='layer1')

    layer2 = Conv2DMaxPool(2, conv_filter_shape=(3, 3), pool_filter_shape=(2, 2), conv_num_filters=128,
                           activation=C.relu, conv_pad=True, pool_strides=(2, 2), name_prefix='layer2')

    layer3 = Conv2DMaxPool(2, conv_filter_shape=(3, 3), pool_filter_shape=(2, 2), conv_num_filters=256,
                           activation=C.relu, conv_pad=True, pool_strides=(2, 2), name_prefix='layer3')

    layer4 = Conv2DMaxPool(3, conv_filter_shape=(3, 3), pool_filter_shape=(2, 2), conv_num_filters=512,
                           activation=C.relu, conv_pad=True, pool_strides=(2, 2), name_prefix='layer4')

    layer5 = Conv2DMaxPool(3, conv_filter_shape=(3, 3), pool_filter_shape=(2, 2), conv_num_filters=512,
                           activation=C.relu, conv_pad=True, pool_strides=(2, 2), name_prefix='layer5')

    dense1 = Dense(4096, activation=C.relu, name='layer6')
    dropout1 = Dropout(0.5)
    dense2 = Dense(4096, activation=C.relu, name='layer7')
    dropout2 = Dropout(0.5)
    dense3 = Dense(num_classes, activation=C.relu, name='layer8')

    def model(x):
        x = layer1(x)
        x = layer2(x)
        x = layer3(x)
        x = layer4(x)
        x = layer5(x)
        x = dropout1(dense1(x))
        x = dropout2(dense2(x))
        x = dense3(x)
        return x

    return model


def VGG19(num_classes: int):
    """ for image classification """
    layer1 = Conv2DMaxPool(2, conv_filter_shape=(3, 3), pool_filter_shape=(2, 2), conv_num_filters=64,
                           activation=C.relu, conv_pad=True, pool_strides=(2, 2), name_prefix='layer1')

    layer2 = Conv2DMaxPool(2, conv_filter_shape=(3, 3), pool_filter_shape=(2, 2), conv_num_filters=128,
                           activation=C.relu, conv_pad=True, pool_strides=(2, 2), name_prefix='layer2')

    layer3 = Conv2DMaxPool(4, conv_filter_shape=(3, 3), pool_filter_shape=(2, 2), conv_num_filters=256,
                           activation=C.relu, conv_pad=True, pool_strides=(2, 2), name_prefix='layer3')

    layer4 = Conv2DMaxPool(4, conv_filter_shape=(3, 3), pool_filter_shape=(2, 2), conv_num_filters=512,
                           activation=C.relu, conv_pad=True, pool_strides=(2, 2), name_prefix='layer4')

    layer5 = Conv2DMaxPool(4, conv_filter_shape=(3, 3), pool_filter_shape=(2, 2), conv_num_filters=512,
                           activation=C.relu, conv_pad=True, pool_strides=(2, 2), name_prefix='layer5')

    dense1 = Dense(4096, activation=C.relu, name='layer6')
    dropout1 = Dropout(0.5)
    dense2 = Dense(4096, activation=C.relu, name='layer7')
    dropout2 = Dropout(0.5)
    dense3 = Dense(num_classes, activation=C.relu, name='layer8')

    def model(x):
        x = layer1(x)
        x = layer2(x)
        x = layer3(x)
        x = layer4(x)
        x = layer5(x)
        x = dropout1(dense1(x))
        x = dropout2(dense2(x))
        x = dense3(x)
        return x

    return model


def conv_bn(input, filter_size, num_filters, strides=(1, 1), init=C.he_normal(), bn_init_scale=1):
    c = Convolution2D(filter_size, num_filters, activation=None, init=init, pad=True, strides=strides, bias=False)(input)
    r = BatchNormalization(map_rank=1, normalization_time_constant=4096, use_cntk_engine=False, init_scale=bn_init_scale, disable_regularization=True)(c)
    return r


def conv_bn_relu(input, filter_size, num_filters, strides=(1, 1), init=C.he_normal()):
    r = conv_bn(input, filter_size, num_filters, strides, init, 1)
    return C.relu(r)


def resnet_bottleneck(input, out_num_filters, inter_out_num_filters):
    c1 = conv_bn_relu(input, (1, 1), inter_out_num_filters)
    c2 = conv_bn_relu(c1, (3, 3), inter_out_num_filters)
    c3 = conv_bn(c2, (1, 1), out_num_filters, bn_init_scale=0)
    p = c3 + input
    return C.relu(p)


def resnet_bottleneck_inc(input, out_num_filters, inter_out_num_filters, stride1x1, stride3x3):
    c1 = conv_bn_relu(input, (1, 1), inter_out_num_filters, strides=stride1x1)
    c2 = conv_bn_relu(c1, (3, 3), inter_out_num_filters, strides=stride3x3)
    c3 = conv_bn(c2, (1, 1), out_num_filters, bn_init_scale=0)
    stride = np.multiply(stride1x1, stride3x3)
    s = conv_bn(input, (1, 1), out_num_filters, strides=stride) # Shortcut
    p = c3 + s
    return C.relu(p)


def resnet_bottleneck_stack(input, num_stack_layers, out_num_filters, inter_out_num_filters):
    assert(num_stack_layers >= 0)
    l = input
    for _ in range(num_stack_layers):
        l = resnet_bottleneck(l, out_num_filters, inter_out_num_filters)
    return l


def create_imagenet_model_bottleneck(input, num_stack_layers, num_classes, stride1x1, stride3x3):
    c_map = [64, 128, 256, 512, 1024, 2048]

    # conv1 and max pooling
    conv1 = conv_bn_relu(input, (7, 7), c_map[0], strides=(2, 2))
    pool1 = MaxPooling((3,3), strides=(2,2), pad=True)(conv1)

    # conv2_x
    r2_1 = resnet_bottleneck_inc(pool1, c_map[2], c_map[0], (1, 1), (1, 1))
    r2_2 = resnet_bottleneck_stack(r2_1, num_stack_layers[0], c_map[2], c_map[0])

    # conv3_x
    r3_1 = resnet_bottleneck_inc(r2_2, c_map[3], c_map[1], stride1x1, stride3x3)
    r3_2 = resnet_bottleneck_stack(r3_1, num_stack_layers[1], c_map[3], c_map[1])

    # conv4_x
    r4_1 = resnet_bottleneck_inc(r3_2, c_map[4], c_map[2], stride1x1, stride3x3)
    r4_2 = resnet_bottleneck_stack(r4_1, num_stack_layers[2], c_map[4], c_map[2])

    # conv5_x
    r5_1 = resnet_bottleneck_inc(r4_2, c_map[5], c_map[3], stride1x1, stride3x3)
    r5_2 = resnet_bottleneck_stack(r5_1, num_stack_layers[3], c_map[5], c_map[3])

    # Global average pooling and output
    pool = AveragePooling(filter_shape=(7, 7), name='final_avg_pooling')(r5_2)
    z = Dense(num_classes, init=C.normal(0.01))(pool)
    return z


def UNET(num_classes, base_num_filters, pad=False):
    """ For semantic segmentation """
    # TODO: allow for depth to be varied
    f = [base_num_filters * 2 ** i for i in range(5)]

    down1 = Conv2DMaxPool(2, conv_filter_shape=(3, 3), pool_filter_shape=(2, 2), conv_num_filters=f[0],
                          activation=C.relu, conv_pad=pad, pool_strides=(2, 2), name_prefix='layer5')

    down2 = Conv2DMaxPool(2, conv_filter_shape=(3, 3), pool_filter_shape=(2, 2), conv_num_filters=f[1],
                          activation=C.relu, conv_pad=pad, pool_strides=(2, 2), name_prefix='layer5')

    down3 = Conv2DMaxPool(2, conv_filter_shape=(3, 3), pool_filter_shape=(2, 2), conv_num_filters=f[2],
                          activation=C.relu, conv_pad=pad, pool_strides=(2, 2), name_prefix='layer5')

    down4 = Conv2DMaxPool(2, conv_filter_shape=(3, 3), pool_filter_shape=(2, 2), conv_num_filters=f[3],
                          activation=C.relu, conv_pad=pad, pool_strides=(2, 2), name_prefix='layer5')

    centre_1 = Convolution2D((3, 3), f[4], activation=C.relu, pad=pad)
    centre_2 = Convolution2D((3, 3), f[4], activation=C.relu, pad=pad)

    up1_1 = Convolution2D((3, 3), f[3], activation=C.relu, pad=pad)
    up1_2 = Convolution2D((3, 3), f[3], activation=C.relu, pad=pad)

    up2_1 = Convolution2D((3, 3), f[2], activation=C.relu, pad=pad)
    up2_2 = Convolution2D((3, 3), f[2], activation=C.relu, pad=pad)

    up3_1 = Convolution2D((3, 3), f[1], activation=C.relu, pad=pad)
    up3_2 = Convolution2D((3, 3), f[1], activation=C.relu, pad=pad)

    up4_1 = Convolution2D((3, 3), f[0], activation=C.relu, pad=pad)
    up4_2 = Convolution2D((3, 3), f[0], activation=C.relu, pad=pad)

    # if num_classes > 1 then softmax is applied at loss function
    clf = Convolution2D((1, 1,), num_classes, activation=identity if num_classes > 1 else C.sigmoid, pad=pad)

    def model(x):
        feature_map0 = x

        # down path
        feature_map1 = down1(feature_map0)
        feature_map2 = down2(feature_map1)
        feature_map3 = down3(feature_map2)
        feature_map4 = down4(feature_map3)

        # latent
        feature_map5 = centre_2(centre_1(feature_map4))

        # up path
        feature_map6 = up1_2(up1_1(Cx.centre_crop_and_splice(feature_map3, Cx.upsample(feature_map5))))
        feature_map7 = up2_2(up2_1(Cx.centre_crop_and_splice(feature_map2, Cx.upsample(feature_map6))))
        feature_map8 = up3_2(up3_1(Cx.centre_crop_and_splice(feature_map1, Cx.upsample(feature_map7))))
        feature_map9 = up4_2(up4_1(Cx.centre_crop_and_splice(feature_map0, Cx.upsample(feature_map8))))

        prediction = clf(feature_map9)
        return prediction

    return model



import cntk as C
import numpy as np
from . import sequence
from . import random
from cntk.layers.blocks import _inject_name


##########################################################################
# linear ops
##########################################################################
def scalar(x, name=''):
    """ select first element of x with shape (1,)

    Arguments:
        x: input tensor

    Returns:
        :class:`~cntk.ops.functions.Function`
        a scalar of shape (1,)
    """
    @C.BlockFunction('scalar', name)
    def inner(x):
        return C.slice(C.reshape(x, (-1,)), 0, 0, 1)

    return inner(x)


def cumsum(x, axis: int = -1):
    """ Calculates the cumulative sum across a static axis

    Arguments:
        x: input tensor
        axis (int): static axis of tensor to cumsum over

    Returns:
        :class:`~cntk.ops.functions.Function`
    """
    d = x.shape[axis]
    u = C.constant(np.triu(np.ones((d, d))).astype(x.dtype))
    if axis != -1:
        x = C.swapaxes(x, -1, axis)
    z = C.times(x, u)
    if axis != -1:
        z = C.swapaxes(z, -1, axis)
    return z


def batchmatmul(left, right, output_rank=1, infer_input_rank_to_map=C.TIMES_NO_INFERRED_INPUT_RANK, name=''):
    """ Batch Matrix Multiplication

    The output of this operation is the matrix product of the two input batch matrices.

    This implementation is similar to tensorflow.matmul.

    Currently assumes the first axis to be the static batch axis. Does not accept multiple static batch axis.

    Example:
        a = C.sequence.input_variable((3, 4, 5))     # batch matrix
        b = C.sequence.input_variable((3, 5, 6))     # batch matrix
        c = Cx.batchmatmul(a, b)
        assert c.shape == (3, 4, 6)                  # 3 is treated as a batch axis


        a = C.sequence.input_variable((3, 4, 5))     # batch matrix
        b = C.sequence.input_variable((3, 5, 6, 7))  # batch tensor
        c = Cx.batchmatmul(a, b, output_rank=2)
        assert c.shape == (3, 4, 6, 7)               # 3 is treated as a batch axis


        a = C.input_variable((3, 4, 5))              # batch matrix
        b = C.input_variable((3, 5, 6, 7))           # batch tensor
        c = Cx.batchmatmul(a, b, output_rank=2)
        assert c.shape == (3, 4, 6, 7)


    Arguments:
        left: left side matrix or tensor
        right: right side matrix or tensor
        output_rank (int): in case we have tensors as arguments, output_rank represents
            the number of axes to be collapsed in order to transform the tensors
            into matrices, perform the operation and then reshape back (explode the axes)
        infer_input_rank_to_map (int): meant for internal use only. Always use default value
        name (str, optional): the name of the Function instance in the network

    Returns:
        :class:`~cntk.ops.functions.Function`
    """

    left_shape = left.shape
    right_shape = right.shape

    seq_axis_present = len(left.dynamic_axes) == 2
    static_batch_axis = left_shape[0]  # assumes the first axis to be the static batch axis.

    if left_shape[0] != right_shape[0]:
        raise ValueError("first axis of left operand and right operand must be the same")

    if (left_shape[0] < 0 or right_shape[0] < 0) and seq_axis_present:
        raise ValueError("Static batch axis cannot be a free axis when dynamic sequence axis is also present")

    # Combine dynamic sequence axis and static batch axis
    if not seq_axis_present:
        left_unpacked = left
        right_unpacked = right
    else:
        left_unpacked = C.sequence.unpack(left, padding_value=0, no_mask_output=True)
        right_unpacked = C.sequence.unpack(right, padding_value=0, no_mask_output=True)

        left_unpacked = C.reshape(left_unpacked, (-1,) + left_shape[1:])
        right_unpacked = C.reshape(right_unpacked, (-1,) + right_shape[1:])

    # Fold static batch axis into dynamic sequence axis
    left_folded = C.to_sequence(left_unpacked)  # do not set sequence length as batch axis has been folded in
    right_folded = C.to_sequence_like(right_unpacked, left_folded)  # seq_length / axis set here to tell cntk they have the same seq axis

    # Matrix Multiply when no static batch axis is present
    result = C.times(left_folded, right_folded, output_rank=output_rank, infer_input_rank_to_map=infer_input_rank_to_map)

    # Split dynamic sequence axis back to original dynamic sequence and static batch axis
    result_unpacked = C.sequence.unpack(result, padding_value=0, no_mask_output=True)
    if not seq_axis_present:
        result_packed = C.reshape(result_unpacked, (static_batch_axis, ) + result.shape)
    else:
        result_unfolded = C.reshape(result_unpacked, (-1, static_batch_axis) + result.shape)
        result_packed = C.to_sequence_like(result_unfolded, left)

    return _inject_name(result_packed, name)


def upsample(x):
    """ Up sample image by a factor of 2 using nearest neighbour.

    Example:
        a = C.input_variable((3, 32, 32)
        b = UpSampling2D(a)

        assert b.shape == (3, 64, 64)

    Arguments:
        x: input image tensor, assumed (channel, row, col)

    Returns:
        :class:`~cntk.ops.functions.Function`

    """
    xr = C.reshape(x, (x.shape[0], x.shape[1], 1, x.shape[2], 1))
    xx = C.splice(xr, xr, axis=-1)  # axis=-1 refers to the last axis
    xy = C.splice(xx, xx, axis=-3)  # axis=-3 refers to the middle axis
    r = C.reshape(xy, (x.shape[0], x.shape[1] * 2, x.shape[2] * 2))
    return r


def centre_crop(larger_image, smaller_image, name: str = ''):
    """ Centre crop spatial dimensions only.

    Arguments:
        larger_image: class:`~cntk.ops.functions.Function` that outputs the tensor to be centre cropped
        smaller_image: class:`~cntk.ops.functions.Function` that outputs the reference tensor
        name (str, optional): the name of the Function instance in the network

    Returns:
        :class:`~cntk.ops.functions.Function`

    """
    input_shape = larger_image.shape  # larger
    referent_shape = smaller_image.shape  # smaller
    row_offset = int((input_shape[1] - referent_shape[1]) / 2)
    col_offset = int((input_shape[2] - referent_shape[2]) / 2)

    if row_offset == 0 and col_offset == 0:
        return larger_image

    elif row_offset < 0 or col_offset < 0:
        raise ValueError(f"offset became negative, check if image was passed correctly. "
                         f"larger image {larger_image.shape}, smaller image {smaller_image.shape}")

    return C.crop_manual(larger_image, smaller_image, row_offset, col_offset, name=name)


def centre_crop_and_splice(larger_image, smaller_image):
    """ Implementation of copy and crop found in UNET architecture.

    Arguments:
        larger_image: to be centre cropped and channel spliced into smaller image
        smaller_image: reference tensor

    Returns:
        :class:`~cntk.ops.functions.Function`

    """
    return C.splice(smaller_image, centre_crop(larger_image, smaller_image), axis=0)


##########################################################################
# non linear and nn ops
##########################################################################
@C.typemap
def swish(x, name=''):
    """ swish activation function first introduced in 'Searching for activation function' by Prajit et al.
    Paper can be found in https://arxiv.org/abs/1710.05941 and https://arxiv.org/abs/1901.02671

    It typically exhibits good performance in a variety of task in vision and nlp problems.
    Can be used as a drop-in replace for relu.
    """

    @C.BlockFunction('Swish', name=name)
    def inner(a):
        return a * C.sigmoid(a)

    return inner(x)


@C.typemap
def hardmax(x, axis=-1, name=''):
    """
    This hardmax implementation can be applied on selected axis. Original cntk hardmax can only be applied on all axis.

    If ``axis`` is given as integer, then the hardmax will be computed along that axis.
    If the provided ``axis`` is -1, it will be computed along the last axis. if None, it will be applied to all axes.

    Arguments:
        x: input_tensor
        axis (int or :class:`~cntk.axis.Axis`): axis along which the hardmax operation will be performed
        name (str, optional): the name of the Function instance in the network

    Returns:
        :class:`~cntk.ops.functions.Function`:
    """

    @C.BlockFunction('Hardmax', name=name)
    def inner(a):
        return C.equal(C.reduce_max(a, axis=axis), a)

    return inner(x)


def erf(x, name=''):
    """
    Computes the element-wise error function of `x`:

    The output tensor has the same shape as ``x``.

    This implementation is from the Handbook of Mathematical Functions and
    has error less than 1.5 * 10-7 for all inputs.
    book can be found here 'http://people.math.sfu.ca/~cbm/aands/frameindex.htm'

    """

    # constants
    a1 = 0.254829592
    a2 = -0.284496736
    a3 = 1.421413741
    a4 = -1.453152027
    a5 = 1.061405429
    p = 0.3275911

    @C.BlockFunction('Erf', name=name)
    def inner(a):
        not_negative = C.greater_equal(a, 0)
        sign = C.element_select(not_negative, not_negative, -1)

        abs_x = C.abs(a)

        # A&S formula 7.1.26
        t = 1.0 / (1.0 + p * a)
        y = 1.0 - (((((a5 * t + a4) * t) + a3) * t + a2) * t + a1) * t * C.exp(-abs_x * abs_x)
        return C.element_times(sign, y)

    return inner(x)


def gelu(x, name=''):
    """ Gaussian Error Linear Unit (GELU), a high-performing neuralnetwork activation function.
    The GELU nonlinearity is the expected transforma-tion of a stochastic regularizer which randomly
    applies the identity or zero mapto a neuron’s input.  The GELU nonlinearity weights inputs by their
    magnitude,rather than gates inputs by their sign as in ReLUs.

    For more detail please refer to 'Gaussian Error Linear Units (GELU)'
    by Hendrycks and Gimpel (https://arxiv.org/abs/1606.08415)

    This activation is used in BERT and OpenAI GPT & GPT-2.

    Its computationally x2 times slower than relu with some negligible increase in memory footprint.

    Arguments:
        x: input_tensor
        name (str, optional): the name of the Function instance in the network

    Returns:
        :class:`~cntk.ops.functions.Function`:

    """
    @C.BlockFunction('Gelu', name=name)
    def inner(a):
        return 0.5 * a * (1 + erf(a / 1.41421356237))

    return inner(x)


def gelu_fast(x, name=''):
    """ This version is an less good approximation of gelu but it is x2 times faster on GPU and x3.8 faster on CPU.
    This implementation just as fast as relu on GPU but x2 slower on CPU.

    Roughly the same memory footprint as relu.

    Arguments:
        x: input_tensor
        name (str, optional): the name of the Function instance in the network

    Returns:
        :class:`~cntk.ops.functions.Function`:

    """
    @C.BlockFunction('GeluFast', name=name)
    def inner(a):
        return a * C.sigmoid(1.702 * a)

    return inner(x)


##########################################################################
# mixture density network ops
##########################################################################
@C.typemap
def gaussian_mdn_coeff(x, nmix: int, ndim: int):
    """
    Extracts the coefficients for gaussian mixture density network.
    Assumes independence between gaussian dimensions.

    Example:
        ndim, nmix = 1, 3
        a = C.input_variable(ndim)
        prediction = Dense((ndim + 2) * nmix)(a)
        coeffs = C.combine(gaussian_mdn_coeff(prediction_tensor, nmix=nmix, ndim=ndim)).eval({a: x})

        alpha, mu, sigma = coeffs.values()

    Arguments:
        x: input tensor
        nmix (int): number of mixture
        ndim (int): number of dimension of gaussian

    Returns:
        tuple

    """

    if len(x.shape) != 1:
        raise ValueError("Must be a 1d tensor, but input has shape {0}".format(x.shape))

    alpha = C.softmax(C.slice(x, 0, 0, nmix), name='alpha')
    sigma = C.exp(C.slice(x, 0, nmix, 2 * nmix), name='sigma')  # common variance for all components in single gaussian kernel
    mu = C.reshape(C.slice(x, 0,  2 * nmix, (ndim + 2) * nmix), shape=(nmix, ndim), name='mu')
    return alpha, mu, sigma


def sample_gaussian_mdn(prediction_tensor, nmix: int, ndim: int):
    """ Constructs sampling nodes from mixture density network outputs

    Example:
        ndim, nmix = 1, 3
        a = C.input_variable(ndim)
        prediction = Dense((ndim + 2) * nmix)(a)
        sampled = sample_gaussian_mdn(prediction, nmix, ndim)

        results = sampled.eval({a: x})  # different results every time you eval

    Arguments:
        prediction_tensor: input tensor
        nmix (int): number of mixture
        ndim (int): number of dimension of gaussian

    Returns:
        :class:`~cntk.ops.functions.Function`

    """
    alpha_tensor, mu_tensor, sigma_tensor = gaussian_mdn_coeff(prediction_tensor, nmix=nmix, ndim=ndim)

    selected_alpha = random.sample(alpha_tensor)
    selected_mu_tensor = C.reduce_sum(mu_tensor * C.expand_dims(selected_alpha, axis=-1), axis=0)
    selected_sigma_tensor = C.reduce_sum(sigma_tensor * selected_alpha, axis=0)

    sampled = C.random.normal_like(selected_sigma_tensor) * selected_sigma_tensor + selected_mu_tensor
    return sampled



'''
First / higher-order functions over sequences, like :func:`Recurrence`.
'''


import cntk as C
import cntkx as Cx
from cntk.ops import splice
from cntk.layers import SentinelValueForAutoSelectRandomSeed
from cntk.layers.blocks import _get_initial_state_or_default, _inject_name
from cntk.default_options import get_default_override, default_override_or
from cntk.layers.sequence import _santize_step_function, RecurrenceFrom


# TODO: Include RecurrenceFrom in here
def Recurrence(step_function, go_backwards=default_override_or(False), initial_state=default_override_or(0),
               return_full_state=False, dropout_rate_input=None,
               dropout_rate_output=None, seed=SentinelValueForAutoSelectRandomSeed, name=''):
    '''
    Recurrence(step_function, go_backwards=False, initial_state=0, return_full_state=False, name='')

    Recurrence has option to variationally dropout input and output.

    Layer factory function that implements a recurrent model, including the common RNN, LSTM, and GRU recurrences.
    This factory function creates a function that runs a step function recurrently over an input sequence,
    where in each step, Recurrence() will pass to the step function a data input as well as the output of the
    previous step.
    The following pseudo-code repesents what happens when you call a `Recurrence()` layer::

      # pseudo-code for y = Recurrence(step_function)(x)
      #  x: input sequence of tensors along the dynamic axis
      #  y: resulting sequence of outputs along the same dynamic axis
      y = []              # result sequence goes here
      s = initial_state   # s = output of previous step ("state")
      for x_n in x:       # pseudo-code for looping over all steps of input sequence along its dynamic axis
          s = step_function(s, x_n)  # pass previous state and new data to step_function -> new state
          y.append(s)

    The common step functions are :func:`~cntk.layers.blocks.LSTM`, :func:`~cntk.layers.blocks.GRU`, and :func:`~cntk.layers.blocks.RNNStep`,
    but the step function can be any :class:`~cntk.ops.functions.Function` or Python function.
    The signature of a step function with a single state variable must be
    ``(h_prev, x) -> h``, where ``h_prev`` is the previous state, ``x`` is the new
    data input, and the output is the new state.
    The step function will be called item by item, resulting in a sequence of the same length as the input.

    Step functions can have more than one state output, e.g. :func:`~cntk.layers.blocks.LSTM`.
    In this case, the first N arguments are the previous state, followed by one more argument that
    is the data input; and its output must be a tuple of N values.
    In this case, the recurrence operation will, by default, return the first of the state variables
    (in the LSTM case, the ``h``), while additional state variables are internal (like the LSTM's ``c``).
    If all state variables should be returned, pass ``return_full_state=True``.

    To provide your own step function, just use any :class:`~cntk.ops.functions.Function` (or equivalent Python function) that
    has a signature as described above.
    For example, a cumulative sum over a sequence can be computed as ``Recurrence(plus)``,
    where each step consists of `plus(s,x_n)`, where `s` is the output of the previous call
    and hence the cumulative sum of all elements up to `x_n`.
    Another example is a GRU layer with projection, which could be realized as ``Recurrence(GRU(500) >> Dense(200))``,
    where the projection is applied to the hidden state as fed back to the next step.
    ``F>>G`` is a short-hand for ``Sequential([F, G])``.

    Optionally, the recurrence can run backwards. This is useful for constructing bidirectional models.

    ``initial_state`` must be a constant. To pass initial_state as a data input, e.g. for a sequence-to-sequence
    model, use :func:`~cntk.layers.sequence.RecurrenceFrom()` instead.

    Note: ``Recurrence()`` is the equivalent to what in functional programming is often called ``scanl()``.

    Example:
     >>> from cntk.layers import Sequential
     >>> from cntk.layers.typing import Tensor, Sequence

     >>> # a recurrent LSTM layer
     >>> lstm_layer = Recurrence(LSTM(500))

     >>> # a bidirectional LSTM layer
     >>> # using function tuples to implement a bidirectional LSTM
     >>> bi_lstm_layer = Sequential([(Recurrence(LSTM(250)),                      # first tuple entry: forward pass
     ...                              Recurrence(LSTM(250), go_backwards=True)),  # second: backward pass
     ...                             splice])                                     # splice both on top of each other
     >>> bi_lstm_layer.update_signature(Sequence[Tensor[13]])
     >>> bi_lstm_layer.shape   # shape reflects concatenation of both output states
     (500,)
     >>> tuple(str(axis.name) for axis in bi_lstm_layer.dynamic_axes)  # (note: str() needed only for Python 2.7)
     ('defaultBatchAxis', 'defaultDynamicAxis')

     >>> # custom step function example: using Recurrence() to
     >>> # compute the cumulative sum over an input sequence
     >>> x = C.input_variable(**Sequence[Tensor[2]])
     >>> x0 = np.array([[   3,    2],
     ...                [  13,   42],
     ...                [-100, +100]])
     >>> cum_sum = Recurrence(C.plus, initial_state=Constant([0, 0.5]))
     >>> y = cum_sum(x)
     >>> y(x0)
     [array([[   3. ,    2.5],
             [  16. ,   44.5],
             [ -84. ,  144.5]], dtype=float32)]

    Args:
     step_function (:class:`~cntk.ops.functions.Function` or equivalent Python function):
      This function must have N+1 inputs and N outputs, where N is the number of state variables
      (typically 1 for GRU and plain RNNs, and 2 for LSTMs).
     go_backwards (bool, defaults to ``False``): if ``True`` then run the recurrence from the end of the sequence to the start.
     initial_state (scalar or tensor without batch dimension; or a tuple thereof):
      the initial value for the state. This can be a constant or a learnable parameter.
      In the latter case, if the step function has more than 1 state variable,
      this parameter must be a tuple providing one initial state for every state variable.
     return_full_state (bool, defaults to ``False``): if ``True`` and the step function has more than one
      state variable, then the layer returns a all state variables (a tuple of sequences);
      whereas if not given or ``False``, only the first state variable is returned to the caller.
     dropout_rate_input (float): dropout for input
     dropout_rate_output (float): dropout for output
     seed (int): seed for randomisation
     name (str, optional): the name of the Function instance in the network

    Returns:
        :class:`~cntk.ops.functions.Function`:
        A function that accepts one argument (which must be a sequence) and performs the recurrent operation on it
    '''

    # BUGBUG: the cum_sum expression in the docstring should be this:
    # cum_sum = Recurrence(C.plus, initial_state=np.array([0, 0.5]))
    # BUGBUG: whereas passing a NumPy array fails with "TypeError: cannot convert value of dictionary"
    # cum_sum = Recurrence(C.plus, initial_state=Constant([0, 0.5]))

    go_backwards  = get_default_override(Recurrence, go_backwards=go_backwards)
    initial_state = get_default_override(Recurrence, initial_state=initial_state)
    initial_state = _get_initial_state_or_default(initial_state)

    step_function = _santize_step_function(step_function)

    dropout_input = None
    if dropout_rate_input:
        dropout_input = VariationalDropout(dropout_rate=dropout_rate_input, seed=seed, name='variational_dropout_input')

    dropout_output = None
    if dropout_rate_output:
        dropout_output = VariationalDropout(dropout_rate=dropout_rate_output, seed=seed, name='variational_dropout_output')

    # get signature of step function
    #*prev_state_args, _ = step_function.signature  # Python 3
    prev_state_args = step_function.signature[0:-1]

    if len(step_function.outputs) != len(prev_state_args):
        raise TypeError('Recurrence: number of state variables inconsistent between create_placeholder() and recurrent block')

    # initial state can be a single value or one per state variable (if more than one, like for LSTM)
    if isinstance(initial_state, tuple) and len(initial_state) == 1:
        initial_state = initial_state[0]
    if not isinstance(initial_state, tuple):
        # TODO: if initial_state is a CNTK Function rather than an initializer, then require to pass it multiple times; otherwise broadcast to all
        initial_state = tuple(initial_state for out_var in prev_state_args)

    # express it w.r.t. RecurrenceFrom
    recurrence_from = RecurrenceFrom(step_function, go_backwards, return_full_state) # :: (x, state seq) -> (new state seq)

    # function that this layer represents
    @C.Function
    def recurrence(x):
        dropped_x = dropout_input(x) if dropout_input else x
        y = recurrence_from(*(initial_state + (dropped_x,)))
        dropped_y = dropout_output(y) if dropout_output else y
        return dropped_y

    return _inject_name(recurrence, name)


def BiRecurrence(step_function: C.Function, initial_state=0, dropout_rate_input=None, dropout_rate_output=None,
                 seed=SentinelValueForAutoSelectRandomSeed, name=''):
    """ Convenience wrapper to create a bidirectional rnn """
    fxn1 = step_function
    fxn2 = step_function.clone(C.CloneMethod.clone, {})

    forward = Recurrence(fxn1, dropout_rate_input=dropout_rate_input, dropout_rate_output=dropout_rate_output, initial_state=initial_state, seed=seed)
    backward = Recurrence(fxn2, dropout_rate_input=dropout_rate_input, dropout_rate_output=dropout_rate_output, initial_state=initial_state, seed=seed, go_backwards=True)

    @C.Function
    def inner(x):
        return C.splice(forward(x), backward(x), axis=-1)

    return inner


def PyramidalBiRecurrence(step_fxn_f, step_fxn_b, width, initial_state_f=0, initial_state_b=0,
                           variational_dropout_rate_input=None, variational_dropout_rate_output=None,
                           seed=SentinelValueForAutoSelectRandomSeed, name=''):
    """ Pyramidal bidirectional recurrence

    Implements a bi-directional recurrence of the step functions given with a pyramidal structure.

    PyramidalBiRecurrence reduces the sequence length and simultaneously increase dimension by width factor.
    Its typically used in acoustic model to keep runtime manageable.

    For mode details, please refer to "Listen, attend and spell" by Chan et al. (https://arxiv.org/abs/1508.01211)

          concatenated
          [f1,b8, f2, b7]
           /    \
          /      \
    <--- b8 <--- b7 <--- b6 <--- b5 <--- b4 <--- b3 <--- b2 <--- b1
    ---> f1 ---> f2 ---> f3 ---> f4 ---> f5 ---> f6 ---> f7 ---> f8

    Example:
        import cntk as C
        import cntkx as Cx

        a = C.sequence.input_variable(10)
        b = Cx.layers.PyramidalBiRecurrence(LSTM(30), LSTM(30), width=2)(a)

        assert b.shape == (30 * 2 * 2, )  # with sequence length also reduced by width factor

    Arguments:
        step_fxn_f: step function in the forward direction
        step_fxn_b: step function in the backward direction
        width (int): width of window in pyramidal structure
        initial_state_f (scalar or tensor without batch dimension; or a tuple thereof):
            the initial value for the state. This can be a constant or a learnable parameter.
            In the latter case, if the step function has more than 1 state variable,
            this parameter must be a tuple providing one initial state for every state variable.
        initial_state_b (scalar or tensor without batch dimension; or a tuple thereof):
            the initial value for the state. This can be a constant or a learnable parameter.
            In the latter case, if the step function has more than 1 state variable,
            this parameter must be a tuple providing one initial state for every state variable.
        variational_dropout_rate_input (float): dropout for input
        variational_dropout_rate_output (float): dropout for output
        seed (int): seed for randomisation
        name (str, optional): the name of the Function instance in the network

    Returns:
        :class:`~cntk.ops.functions.Function`:

    """
    forward = Recurrence(step_fxn_f, initial_state=initial_state_f, go_backwards=False,
                         return_full_state=False, dropout_rate_input=variational_dropout_rate_input,
                         dropout_rate_output=variational_dropout_rate_output, seed=seed, name='Pyramidal_Forward')

    backward = Recurrence(step_fxn_b, initial_state=initial_state_b, go_backwards=True,
                          return_full_state=False, dropout_rate_input=variational_dropout_rate_input,
                          dropout_rate_output=variational_dropout_rate_output, seed=seed, name='Pyramidal_Backward')

    @C.Function
    def inner(x):
        y = C.splice(forward(x), backward(x), axis=-1)
        z = Cx.sequence.window(y, width)
        return z

    return inner


def VariationalDropout(dropout_rate: float, seed=SentinelValueForAutoSelectRandomSeed, name=''):
    """ Variational dropout uses the same dropout mask at each time step (i.e. across the dynamic sequence axis)

    Example:
        a = C.sequence.input_variable(10)
        b = VariationalDropout(0.1)(a)

        assert b.shape == a.shape

    Arguments:
        dropout_rate (float): probability of dropping out an element
        seed (int): seed for randomisation
        name (str, defaults to ''): the name of the Function instance in the network

    Returns:
        cntk.ops.functions.Function:
        A function that accepts one argument and applies the operation to it

    """
    dropout = C.layers.Dropout(dropout_rate, seed=seed)

    @C.BlockFunction('VariationalDropout', name)
    def inner(x):
        mask = dropout(C.ones_like(C.sequence.first(x)))
        mask = C.sequence.broadcast_as(mask, x)
        return mask * x

    return inner



import cntk as C
import math
import numpy as np
from typing import List


class CyclicalLearningRate(object):
    """
    Cyclical learning rate is an implementation to that  practically eliminates
    the need to experimentally find the best values and schedule  for  the
    global  learning  rates.

    Instead  of  monotonically decreasing the learning rate, this method lets the
    learning  rate  cyclically  vary  between  reasonable  boundary  values

    Training  with  cyclical  learning  rates  instead of  fixed  values  achieves
    improved  classification  accuracy without a need to tune and often in fewer iterations.

    This is an CNTK implementation of the following paper:
    Cyclical Learning Rates for Training Neural Networks by Leslie N. Smith: https://arxiv.org/abs/1506.01186

    The policy cycles the learning rate between two boundaries with a constant frequency, as detailed in
    the paper.
    The distance between the two boundaries can be scaled on a per-iteration
    or per-cycle basis.

    Cyclical learning rate policy changes the learning rate after every batch.
    `batch_step` should be called after a batch has been used for training.
    To resume training, save `last_batch_iteration` and use it to instantiate `CyclicalLeaningRate`.

    This class has three built-in policies, as put forth in the paper:
    "triangular":
        A basic triangular cycle w/ no amplitude scaling.
    "triangular2":
        A basic triangular cycle that scales initial amplitude by half each cycle.
    "exp_range":
        A cycle that scales initial amplitude by gamma**(cycle iterations) at each
        cycle iteration.

    This implementation was adapted from the github repo:
    `bckenstler/CLR` and 'anandsaha/pytorch.cyclic.learning.rate'


    Args:
        parameter_learner (learner): list of cntk learner
        base_lr (float or list): Initial learning rate which is the
            lower boundary in the cycle for eachparam groups.
            Default: 0.001
        max_lr (float or list): Upper boundaries in the cycle for
            each parameter group. Functionally,
            it defines the cycle amplitude (max_lr - base_lr).
            The lr at any cycle is the sum of base_lr
            and some scaling of the amplitude; therefore
            max_lr may not actually be reached depending on
            scaling function. Default: 0.006
        ramp_up_step_size (int): Number of training iterations in the
            lr ramp up phase. Authors suggest setting step_size
            2-8 x training iterations in epoch. Default: 2000
        ramp_down_step_size (int): Number of training iterations in the
            lr ramp down phase. If not set, it will take the value of
            ramp_up_step_size making the lr curve symmetric.
        minibatch_size (int): Number of samples in one minibatch
        lr_policy (str): One of {triangular, triangular2, exp_range}.
            Values correspond to policies detailed above.
            If scale_fn is not None, this argument is ignored.
            Default: 'triangular2'
        gamma (float): Constant in 'exp_range' scaling function:
            gamma ** iterations
            Default: 0.99994
        scale_fn (function): Custom scaling policy defined by a single
            argument lambda function, where
            0 <= scale_fn(x) <= 1 for all x >= 0.
            lr_policy is ignored
            Default: None
        scale_by (str): scale by either number of training iterations or training cycles.
            Only used if custom scaling policy scale_fn is defined.
        record_history (bool): If True, loss & learn rate will be recorded. get loss lr
            history functions can then be used.
        last_batch_iteration (int): The index of the last batch. Default: -1
    Example:
     >>> model = C.layers.Dense(10)(C.input_variable(10))
     >>> sgd_momentum = C.momentum_sgd(model.parameters, 0.1, 0.9)
     >>> clr = CyclicalLeaningRate(sgd_momentum, minibatch_size=32)

     >>> for epoch in range(10):
     ...     for batch in range(100):
     ...         # trainer.train_minibatch(...)
     ...         clr.batch_step()  # must be called once for every training iteration/update
    """

    def __init__(self, parameter_learner, base_lr=1e-3, max_lr=6e-3, minibatch_size=None,
                 ramp_up_step_size=2000, ramp_down_step_size: int = None, lr_policy: str = 'triangular2', gamma=0.99994,
                 scale_fn=None, scale_by: str = None, record_history: bool = False, last_batch_iteration: int = -1):

        if lr_policy not in ['triangular', 'triangular2', 'exp_range'] and scale_fn is None:
            raise ValueError('lr_policy is invalid and scale_fn is None')

        if scale_by is not None and scale_by not in ['iteration', 'cycle']:
            raise ValueError("Can only scale by iteration or cycle")

        if scale_by is not None and scale_fn is None:
            raise ValueError('scale_by can only be used when custom scale function is used')

        self.parameter_learner = parameter_learner
        self.base_lr = base_lr
        self.max_lr = max_lr

        self.ramp_up_step_size = ramp_up_step_size
        self.ramp_down_step_size = ramp_down_step_size or ramp_up_step_size
        self.cycle_size = self.ramp_up_step_size + self.ramp_down_step_size
        self.minibatch_size = minibatch_size

        self.lr_policy = lr_policy
        self.gamma = gamma
        self.record_history = record_history

        if scale_fn is None:
            if self.lr_policy == 'triangular':
                self.scale_fn = self._triangular_scale_fn
                self.scale_by = 'cycle'
            elif self.lr_policy == 'triangular2':
                self.scale_fn = self._triangular2_scale_fn
                self.scale_by = 'cycle'
            elif self.lr_policy == 'exp_range':
                self.scale_fn = self._exp_range_scale_fn
                self.scale_by = 'iteration'
        else:
            self.scale_fn = scale_fn
            self.scale_by = scale_by

        self.loss = []
        self.lrs: List[float] = []  # [(learn1_lr, ), (learner1_lr, ), ... ]
        self.current_lr = 0
        self.last_batch_iteration = last_batch_iteration
        self.batch_step()

    def _triangular_scale_fn(self, x) -> float:
        return 1.

    def _triangular2_scale_fn(self, x) -> float:
        return 1 / (2. ** (x - 1))

    def _exp_range_scale_fn(self, x) -> float:
        return self.gamma ** x

    def batch_step(self, previous_minibatch_loss=None):
        """
        Updates learners with new learning rate after one training iteration is complete.
        Must be called once for every training iteration/update.
        """

        self.last_batch_iteration += 1
        lr = self.get_lr()
        self.current_lr = lr
        
        # loss and learn rate gets recorded in pre-training mode
        if self.record_history and previous_minibatch_loss:
            self.loss.append(previous_minibatch_loss)
            self.lrs.append(lr)

        self.parameter_learner.reset_learning_rate(C.learning_parameter_schedule(lr, minibatch_size=self.minibatch_size))
        return None

    def get_lr(self) -> float:
        """ Get learning rate based on last_batch_iteration count """
        # Cycle number
        cycle_num: int = math.floor(1 + self.last_batch_iteration / self.cycle_size)

        # number of batch steps made since last complete cycle
        iteration_since_last_cycle = self.last_batch_iteration % self.cycle_size

        # ramping up or down
        is_ramp_up = True if iteration_since_last_cycle <= self.ramp_up_step_size else False

        num_batch_steps_in_ramp = iteration_since_last_cycle
        if is_ramp_up:
            base_height = (self.max_lr - self.base_lr) * (num_batch_steps_in_ramp / self.ramp_up_step_size)
        else:  # ramp_down
            num_batch_steps_in_ramp = iteration_since_last_cycle - self.ramp_up_step_size
            base_height = (self.max_lr - self.base_lr) * (1 - num_batch_steps_in_ramp / self.ramp_down_step_size)

        # base_height = (self.max_lr - self.base_lr) * max(0., (1 - proportion_of_step_completed))
        xx = cycle_num if self.scale_by == "cycle" else self.last_batch_iteration
        lr = self.base_lr + base_height * self.scale_fn(xx)
        return lr

    def get_lr_schedule(self, number_of_cycles=4) -> np.ndarray:
        """ returns how the learn rate schedule will be like. Useful to check if your
        custom learning policy is working as intended.

        The returned lr_schedule can be visualised in the following:
            import matplotlib.pyplot as plt

            plt.scatter(range(lr_schedule.shape[0]), lr_schedule, s=1)
            plt.show()

        Arguments:
            number_of_cycles (int):

        Returns:
            np.ndarray 1-d

        """
        store_last_iteration = self.last_batch_iteration
        self.last_batch_iteration = 0

        lr_schedule = []
        for i in range(number_of_cycles * self.cycle_size):
            lr_schedule.append(self.get_lr())
            self.last_batch_iteration += 1

        self.last_batch_iteration = store_last_iteration
        lr_schedule = np.array(lr_schedule)
        assert lr_schedule.ndim == 1
        return lr_schedule

    def get_loss_lr_history(self) -> np.ndarray:
        assert self.record_history, "Cannot be used outside of pre-training as loss is not captured"
        return np.array([(loss, *lrs) for loss, lrs in zip(self.loss, self.lrs)])

    def get_averaged_loss_lr_history(self, window=100) -> np.ndarray:
        """ Average loss and learn rate value within window size

        Any remainder outside of window size will not be included in the average returned.

        Mean values returned can be visualised in the follow:

            import matplotlib.pyplot as plt
            plt.scatter(mean_loss_lr[1], mean_loss_lr[0], s=1)
            plt.show()

        Using the graph, determine the base_lr and max_lr.

        Base_lr is the smallest lr value that results in loss decreasing.
        Max_lr is the largest lr before loss becomes unstable.

        """
        assert self.record_history, "Cannot be used outside of pre-training as loss is not captured"
        assert window > 0, "window size cannot be zero or smaller"

        history = self.get_loss_lr_history()

        if window > 1:
            remainder = history.shape[0] % window
            if remainder:
                print(f"last {remainder} rows are exlcuded from average calculation")
                history = history[:-remainder, ...]

            history = history.reshape((-1, window, history.shape[1]))
            history = np.mean(history, axis=1)

        return history



from .. import CTCEncoder
import cntk as C
from cntk.layers import Dense, LSTM, Recurrence
import numpy as np


def test_ctc_encoder_string_labels():
    label_set = list('abc')
    encoder = CTCEncoder(label_set)

    seq_length = 7
    labels = list('abbbc')
    actual = encoder.transform(labels, seq_length)

    desired = [[0, 2, 0, 0],
               [0, 0, 2, 0],
               [0, 0, 1, 0],
               [0, 0, 2, 0],
               [0, 0, 1, 0],
               [0, 0, 2, 0],
               [0, 0, 0, 2],]

    desired = np.array(desired)

    np.testing.assert_equal(actual, desired)

    # test inverse
    decoded = encoder.inverse_transform(actual)
    assert labels == decoded

    seq_length = 10
    labels = list('abbbc')
    actual = encoder.transform(labels, seq_length)

    desired = [[0, 2, 0, 0],
               [0, 0, 2, 0],
               [0, 0, 1, 0],
               [0, 0, 2, 0],
               [0, 0, 1, 0],
               [0, 0, 2, 0],
               [0, 0, 0, 2],
               [0, 0, 0, 1],
               [0, 0, 0, 1],
               [0, 0, 0, 1],]

    desired = np.array(desired)

    np.testing.assert_equal(actual, desired)

    # test inverse
    decoded = encoder.inverse_transform(actual)
    assert labels == decoded

    seq_length = 10
    labels = list('abbba')
    actual = encoder.transform(labels, seq_length)

    desired = [[0, 2, 0, 0],
               [0, 0, 2, 0],
               [0, 0, 1, 0],
               [0, 0, 2, 0],
               [0, 0, 1, 0],
               [0, 0, 2, 0],
               [0, 2, 0, 0],
               [0, 1, 0, 0],
               [0, 1, 0, 0],
               [0, 1, 0, 0], ]

    desired = np.array(desired)

    np.testing.assert_equal(actual, desired)

    # test inverse
    decoded = encoder.inverse_transform(actual)
    assert labels == decoded

    seq_length = 15
    labels = list('abcabcabc')
    actual = encoder.transform(labels, seq_length)

    desired = [[0, 2, 0, 0],
               [0, 0, 2, 0],
               [0, 0, 0, 2],
               [0, 2, 0, 0],
               [0, 0, 2, 0],
               [0, 0, 0, 2],
               [0, 2, 0, 0],
               [0, 0, 2, 0],
               [0, 0, 0, 2],]

    desired += [[0, 0, 0, 1]] * 6

    desired = np.array(desired)

    np.testing.assert_equal(actual, desired)

    # test inverse
    decoded = encoder.inverse_transform(actual)
    assert labels == decoded


def test_ctc_encoder_int_labels():
    label_set = list(range(3))
    encoder = CTCEncoder(label_set)

    seq_length = 7
    labels = [1, 2, 2, 2, 3]
    actual = encoder.transform(labels, seq_length)

    desired = [[0, 2, 0, 0],
               [0, 0, 2, 0],
               [0, 0, 1, 0],
               [0, 0, 2, 0],
               [0, 0, 1, 0],
               [0, 0, 2, 0],
               [0, 0, 0, 2],]

    desired = np.array(desired)

    np.testing.assert_equal(actual, desired)

    # test inverse
    decoded = encoder.inverse_transform(actual)
    assert labels == decoded

    seq_length = 10
    labels = [1, 2, 2, 2, 3]
    actual = encoder.transform(labels, seq_length)

    desired = [[0, 2, 0, 0],
               [0, 0, 2, 0],
               [0, 0, 1, 0],
               [0, 0, 2, 0],
               [0, 0, 1, 0],
               [0, 0, 2, 0],
               [0, 0, 0, 2],
               [0, 0, 0, 1],
               [0, 0, 0, 1],
               [0, 0, 0, 1],]

    desired = np.array(desired)

    np.testing.assert_equal(actual, desired)

    # test inverse
    decoded = encoder.inverse_transform(actual)
    assert labels == decoded

    seq_length = 10
    labels = [1, 2, 2, 2, 1]
    actual = encoder.transform(labels, seq_length)

    desired = [[0, 2, 0, 0],
               [0, 0, 2, 0],
               [0, 0, 1, 0],
               [0, 0, 2, 0],
               [0, 0, 1, 0],
               [0, 0, 2, 0],
               [0, 2, 0, 0],
               [0, 1, 0, 0],
               [0, 1, 0, 0],
               [0, 1, 0, 0], ]

    desired = np.array(desired)

    np.testing.assert_equal(actual, desired)

    # test inverse
    decoded = encoder.inverse_transform(actual)
    assert labels == decoded


def test_ctc_encoder_train_and_network_output_to_labels():
    # test CTC encoder in training loop and CTCEncoder.network_output_to_labels

    a = C.sequence.input_variable(10)
    labels = ['a', 'b', 'c']
    encoder = CTCEncoder(labels)

    labels_tensor = C.sequence.input_variable(len(encoder.classes_))  # number of classes = 4
    input_tensor = C.sequence.input_variable(100)

    prediction_tensor = Dense(4)(Recurrence(LSTM(100))(C.ones_like(input_tensor)))

    labels_graph = C.labels_to_graph(labels_tensor)

    fb = C.forward_backward(labels_graph, prediction_tensor, blankTokenId=encoder.blankTokenId)

    ground_truth = ['a', 'b', 'b', 'b', 'c']
    seq_length = 10  # must be the same length as the sequence length in network_out

    pred = np.array([[0., 2., 0., 0.],
                     [0., 2., 0., 0.],
                     [0., 0., 2., 0.],
                     [2., 0., 0., 0.],
                     [0., 0., 2., 0.],
                     [2., 0., 0., 0.],
                     [0., 0., 2., 0.],
                     [2., 0., 0., 0.],
                     [0., 0., 0., 2.],
                     [0., 0., 0., 2.], ]).astype(np.float32)

    n = np.random.random((10, 100)).astype(np.float32)

    # result = fb.eval({labels_tensor: [encoder.transform(ground_truth, seq_length=seq_length)],
    #                   input_tensor: [n]})

    # print(result)

    adam = C.adam(prediction_tensor.parameters, 0.01, 0.912)
    trainer = C.Trainer(prediction_tensor, (fb,), [adam])

    for i in range(300):
        trainer.train_minibatch({labels_tensor: [encoder.transform(ground_truth, seq_length=seq_length)],
                                 input_tensor: [n]})

        # print(trainer.previous_minibatch_loss_average)

    result = prediction_tensor.eval({input_tensor: [n]})
    assert encoder.network_output_to_labels(result[0], squash_repeat=True) == ground_truth



import cntk as C
from cntkx.ops.sequence import length, pad, stride, position, join, window, reverse, reduce_mean
import numpy as np


def test_sequence_length():
    s1 = 5
    s2 = 3
    a = C.sequence.input_variable(1)
    b = length(a)
    n1 = np.random.random((s1, 1)).astype(np.float32)
    n2 = np.random.random((s2, 1)).astype(np.float32)

    result = b.eval({a: [n1, n2]})
    assert result[0][0] == s1, result[0][0]
    assert result[1][0] == s2, result[0][0]


def test_position():
    a = C.sequence.input_variable((1, 2, 3, 4, 5))
    p = position(a)

    n = [np.random.random((4, 1, 2, 3, 4, 5)).astype(np.float32),
         np.random.random((11, 1, 2, 3, 4, 5)).astype(np.float32),
         np.random.random((15, 1, 2, 3, 4, 5)).astype(np.float32)]

    results = p.eval({a: n})

    for actual, nn in zip(results, n):
        desired = np.arange(nn.shape[0])
        np.testing.assert_equal(actual, desired)


def test_sequence_pad():
    pattern = (1, 1)
    input_dim = 10
    s1 = 5
    s2 = 3
    a = C.sequence.input_variable(input_dim)
    b = pad(a, pattern)
    c = length(b)
    n1 = np.random.random((s1, input_dim)).astype(np.float32)
    n2 = np.random.random((s2, input_dim)).astype(np.float32)

    result = c.eval({a: [n1, n2]})
    assert result[0][0] == s1 + sum(pattern), result[0][0]
    assert result[1][0] == s2 + sum(pattern), result[0][0]
    assert b.shape == (input_dim, )


def test_sequence_pad1():
    pattern = (2, 2)
    input_dim = 10
    s1 = 5
    s2 = 3
    a = C.sequence.input_variable(input_dim)
    b = pad(a, pattern)
    c = length(b)
    n1 = np.random.random((s1, input_dim)).astype(np.float32)
    n2 = np.random.random((s2, input_dim)).astype(np.float32)

    result = c.eval({a: [n1, n2]})
    assert result[0][0] == s1 + sum(pattern), result[0][0]
    assert result[1][0] == s2 + sum(pattern), result[0][0]
    assert b.shape == (input_dim, )


def test_stride():
    contexts = [(10, 2), (10, 3), (10, 4), (10, 5), (10, 6),
                (5000, 2), (5000, 3), (5000, 4), (5000, 5), (5000, 6),
                (1000000, 2)]  # [(seq_length, stride), ...]

    a = C.sequence.input_variable(10)

    for seq_length, s in contexts:

        b = stride(a, s)

        # n = np.arange(seq_length * 1).reshape((1, seq_length, 1)).astype(np.float32)
        n = np.random.random((1, seq_length, 10)).astype(np.float32)

        output = b.eval({a: n})[0]
        np.testing.assert_equal(output, n[0][::s, ...])

    # test with mini-batch
    s = 2
    n = [np.random.random((10, 10)).astype(np.float32),
         np.random.random((5000, 10)).astype(np.float32),
         np.random.random((10000, 10)).astype(np.float32)]

    b = stride(a, s)

    output = b.eval({a: n})[-1]
    np.testing.assert_equal(output, n[-1][::s, ...])

    # test with sequence of rgb images
    a = C.sequence.input_variable((3, 64, 64))
    s = 2
    n = [np.random.random((6, 3, 64, 64)).astype(np.float32),
         np.random.random((11, 3, 64, 64)).astype(np.float32),
         np.random.random((15, 3, 64, 64)).astype(np.float32)]

    b = stride(a, s)

    output = b.eval({a: n})

    for actual, desired in zip(output, n):
        np.testing.assert_equal(actual, desired[::s])


def test_join():
    a = C.sequence.input_variable(3)
    b = C.sequence.input_variable(3)

    ab = join(a, b)

    n = [np.random.random((2, 3)).astype(np.float32),
         np.random.random((4, 3)).astype(np.float32),
         np.random.random((6, 3)).astype(np.float32), ]

    m = [np.random.random((2, 3)).astype(np.float32),
         np.random.random((4, 3)).astype(np.float32),
         np.random.random((6, 3)).astype(np.float32), ]

    joined = ab.eval({a: n, b: m})
    nm = [np.concatenate((nn, mm), axis=0) for nn, mm in zip(n, m)]

    for result, desired in zip(joined, nm):
        np.testing.assert_equal(result, desired)


def test_window():
    # ====================================================================
    # window width = 2
    # ====================================================================
    seq_length = 20
    dim = 1
    k = 2
    a = C.sequence.input_variable(dim)
    b = window(a, width=k)

    assert b.shape == (dim * k, )

    n = np.random.random((1, seq_length, dim)).astype(np.float32)
    m = np.pad(n[:, 1:, :], ((0, 0), (0, 1), (0, 0)), mode='constant', constant_values=0,)

    result = b.eval({a: n})[0]
    desired = np.concatenate((n, m), axis=-1)[0, ::k]

    np.testing.assert_equal(result, desired)

    # ====================================================================
    # window width = 3
    # ====================================================================
    seq_length = 16
    dim = 4
    k = 3
    a = C.sequence.input_variable(dim)
    b = window(a, width=k)

    assert b.shape == (dim * k,)

    n = np.random.random((1, seq_length, dim)).astype(np.float32)
    m = np.pad(n[:, 1:, :], ((0, 0), (0, 1), (0, 0)), mode='constant', constant_values=0, )
    o = np.pad(n[:, 2:, :], ((0, 0), (0, 2), (0, 0)), mode='constant', constant_values=0, )

    result = b.eval({a: n})[0]
    desired = np.concatenate((n, m, o), axis=-1)[0, ::k]

    np.testing.assert_equal(result, desired)


def test_reverse():
    ndim = 3
    a = C.sequence.input_variable(ndim)
    r = reverse(a)

    n = [np.arange(10 * ndim).reshape((10, ndim)).astype(np.float32),
         np.arange(8 * ndim).reshape((8, ndim)).astype(np.float32),
         np.arange(5 * ndim).reshape((5, ndim)).astype(np.float32), ]

    results = r.eval({a: n})

    for result, input_array in zip(results, n):
        desired = input_array[::-1, ...]
        np.testing.assert_equal(result, desired)


def test_reduce_mean():
    a = C.sequence.input_variable(32)
    b = reduce_mean(a)

    n = [np.random.random((10, 32)).astype(np.float32),
         np.random.random((10, 32)).astype(np.float32),
         np.random.random((10, 32)).astype(np.float32),]

    results = b.eval({a: n})

    for r, d in zip(results, n):
        np.testing.assert_almost_equal(r, np.mean(d, axis=0))

    a = C.sequence.input_variable((3, 4))
    b = reduce_mean(a)

    n = [np.random.random((10, 3, 4)).astype(np.float32),
         np.random.random((10, 3, 4)).astype(np.float32),
         np.random.random((10, 3, 4)).astype(np.float32),]

    results = b.eval({a: n})

    for r, d in zip(results, n):
        np.testing.assert_almost_equal(r, np.mean(d, axis=0))



import cntk as C


# @C.typemap
def sample(x, axis=-1, name=''):
    """ Sample an unnormalised log-prob distribution and returns a one hot encode vector

    Arguments:
        x: input tensor
        name (str): the name of the Function instance in the network

    Returns:
        :class:`~cntk.ops.functions.Function`

    """

    @C.BlockFunction('Random::Sample', name=name)
    def inner(a):
        perturbed = a + C.random.gumbel_like(a)
        sampled = C.equal(C.reduce_max(perturbed, axis=axis), perturbed, name=name)  # equivalent to hardmax(perturbed_x)
        return sampled

    return inner(x)


# @C.typemap
def sample_top_k(x, k, num_classes, axis=-1, name=''):
    """ Sample once from the top_k unnormalised log-prob distribution of `x` and returns a one hot encoded vector.

    Example:
        import cntk as C
        import cntkx as Cx

        a = C.input_variable(5)
        b = Cx.random.sample_top_k(a, k=3, num_classes=5)

        n = np.array([[1, 2, 3, 4, 5],] * 1000)

        results = b.eval({a: n})
        assert np.sum(results[:, :2]) == 0
        assert np.sum(results[:, 2:]) == 1000


    Arguments:
        x: input tensor
        k (int): number of k largest probability to sample from
        num_classes (int): typically the dimension of `x`
        axis (int): axis along which to perform the operation (default: -1)
        name (str): the name of the Function instance in the network

    Returns:
        :class:`~cntk.ops.functions.Function`

    """

    @C.BlockFunction('Random::SampleTopK', name=name)
    def inner(a):
        # a: [#, *] [static_axes, num_classes]

        k_values, k_indices = C.top_k(a, k=k, axis=axis).outputs
        # k_indices [#, *] [static_axes, k]

        b = C.one_hot(k_indices, num_classes)
        # b: [#, *] [static_axes, k, num_classes]

        valid_probabilities = C.squeeze(C.reduce_sum(b, axis=-2), axes=(-2,))
        # valid_probabilities: [#, *] [static_axes, num_classes]

        # k largest probabilies are retained, everything else is set to -inf and will not be sampled
        minus_inf = C.constant(-1e+30)
        d = a * valid_probabilities
        e = C.element_select(d, d, minus_inf)
        # e: [#, *] [static_axes, num_classes]

        # sample from top_k distribution once
        s = sample(e, axis=axis, name=name)
        # s: [#, *] [static_axes, num_classes]
        return s

    return inner(x)
